#=======================================================================
#	$Id: WebFiler.pm,v 1.6 2011/02/14 16:38:08 pythontech Exp $
#	Serve files via http (simpler and less restrictive than DAV).
#=======================================================================
package PythonTech::WebFiler;
use strict;
use HTTP::Date;
use File::Path;
use PythonTech::Conv qw(uq hq);

#-----------------------------------------------------------------------
#	Create new server object
#	Configuration options:
#	  basePath		Root of filestore to be exported
#	  authentication	How to authenticate remote user:
#		none		Allow anybody access (unsafe!)
#		htaccess	Assume script is protected by .htaccess
#				rules e.g. "require valid-user" or
#				"require group foobar"
#		cookie		User is associated with session cookie
#				using PythonTech::WebServer framework
#	  readPermission	Who has read permission.  Can be a Perl
#				list of usernames, or '*'
#	  writePermission	Who has write permission.  Can be a Perl
#				list of usernames, or '*'
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {@_};
    foreach (qw(basePath)) {
	die "$class: $_ not configured\n" unless $self->{$_};
    }
    $self->{basePath} .= '/'
	unless substr($self->{basePath},-1) eq '/';
    bless $self, $class;
    return $self;
}

sub webquery {
    my($self, $query) = @_;
    unless (eval {$self->_webquery($query); 1;}) {
	my $exc = $@;
	print STDERR "Exception: ###$exc###\n";
	my($exc1) = $exc =~ /^(.*)/; # First line only
	my $resp = $query->response;
	$resp->set_status("500 $exc1");
	$resp->set_type('text/plain');
	$resp->write($exc);
    }
}

sub _webquery {
    my($self, $query) = @_;
    my $user;
    my $authtype = $self->{'authentication'};
    if ($authtype eq 'htaccess') {
	# Assume protected by .htaccess rule
	$user = $query->remote_user;
    } elsif ($authtype eq 'none') {
	# Unsafe!
	$user = '';
    } elsif ($authtype eq 'cookie') {
	# Crude authorisation: authentication
	$user = $query->require_login
	    or return;
    } else {
	die "Unknown authentication type: $authtype\n";
    }
    my $response = $query->response;
    my $method = $query->request_method;

    # All requests have a path.  Check it is valid.
    my $tpath = $query->param('path'); # Tainted
    my($path) = $tpath =~ m!^([^\0]*)$!
	or die "Invalid NUL in path\n";
    die "'..' not allowed\n" if "/$tpath/" =~ m!/\.\./!;
    my $fullpath = $self->{'basePath'} . $path;

    my $action = $query->param('action');
    if ($method eq 'GET') {
	$self->permit($query, $user, 'read')
	    or return;
	if (! defined $action) {
	    # Read whatever type it is
	    my $type = &check_type($fullpath, '','d','f','l');
	    if ($type eq 'd') {
		$self->_readdir($response, $fullpath);
	    } elsif ($type eq 'f') {
		$self->_readfile($response, $fullpath);
	    } elsif ($type eq 'l') {
		$self->_readlink($response, $fullpath);
	    } else {
		if ($response->can('not_found')) {
		    $response->not_found();
		} else {
		    # Backward compatibility
		    my $q = $response->{'cgi'};	# FIXME naughty - need 404 method
		    print $q->header(-status=>404);
		    print "<title>Not found</title>$fullpath not found";
		}
	    }
	# Otherwise, client claims they know what type of file it is
	} elsif ($action eq 'readfile') {
	    &check_type($fullpath, 'f');
	    $self->_readfile($response, $fullpath);
	} elsif ($action eq 'readlink') {
	    &check_type($fullpath, 'l');
	    $self->_readlink($response, $fullpath);
	} elsif ($action eq 'readdir') {
	    &check_type($fullpath, 'd');
	    $self->_readdir($response, $fullpath);
	} else {
	    die "Action $action unknown\n";
	}
    } elsif ($method eq 'POST') {
	$self->permit($query, $user, 'write')
	    or return;
	(my $dir = $fullpath) =~ s![^/]+$!!;
	my $leaf = substr($fullpath,length($dir));
	if ($action eq 'write') {
	    # Should be PUT method perhaps?
	    &check_dir_and_type($dir,$leaf, '', 'f');
	    my $fh = $query->upload('content')
		or die "No content file uploaded\n";
	    $self->_writefile($dir,$leaf,$fh,
			      0 + $query->param('mtime'),
			      $query->param('mode'));
	    $self->no_content($response);
	} elsif ($action eq 'mkdir') {
	    &check_dir_and_type($dir,$leaf, '');
	    $self->_mkdir($dir, $leaf, $query->param('mode'));
	    $self->no_content($response);
	} elsif ($action eq 'writelink') {
	    my $link = $query->param('link')
		or die "link not supplied\n";
	    &check_dir_and_type($dir,$leaf, '', 'l', 'f');
	    $self->_writelink($dir, $leaf, $link);
	    $self->no_content($response);
	} else {
	    die "Action $action unknown\n";
	}
    } else {
	die "Method $method not supported\n";
    }
}

#-----------------------------------------------------------------------
#	Low-level methods, after access and validity checks have been
#	done.  Filesystem-level access problems will raise exceptions.
#-----------------------------------------------------------------------
sub _readfile {
    my($self, $response,$fullpath) = @_;
    my $mtime = (stat $fullpath)[9];
    open(PTWFILE,'<',$fullpath) || 
	die "Cannot open $fullpath for read: $!\n";
    binmode(PTWFILE);
    $response->set_type("application/octet-stream");
    $response->set_header(last_modified=>&HTTP::Date::time2str($mtime));
    my $data;
    while (read(PTWFILE,$data,4096)) {
	$response->write($data);
    }
    $response->write(''); # Flush, in case zero length file
    close(PTWFILE);
}

sub _readlink {
    my($self, $response,$fullpath) = @_;
    my $link = readlink($fullpath);
    $response->set_type('application/x-ptwb-symlink');
    $response->write($link);
}

sub _readdir {
    my($self, $response,$fullpath) = @_;
    my $listing;
    opendir(PTWDIR,$fullpath) ||
	die "Cannot open directory $fullpath for read: $!\n";
    foreach my $leaf (readdir(PTWDIR)) {
	next if $leaf eq '.' || $leaf eq '..';
	next if substr($leaf,0,10) eq '.webfiler.'; # Our temp files
	my $filename = "$fullpath/$leaf";
	my @stat = lstat($filename);
	$listing .= sprintf("%s %s %o %d %d %s\n",
			    -l _ ? "l" :
			    -d _ ? "d" :
			    -f _ ? "f" : "?",
			    uq($leaf),
			    $stat[2], # Mode
			    $stat[7], # Size
			    $stat[9], # Modtime
			    -l _ ? uq(readlink($filename)) : "");
    }
    closedir(PTWDIR);
    $response->set_type("application/x-ptwb-dir");
    $response->write($listing);
}

sub _writefile {
    my($self, $dir,$leaf,$fh,$mtime,$mode) = @_;
    my $dest = "$dir$leaf";
    my $newpath = $self->tempname($dir);
    eval {
	open(PTWFILE,'>',$newpath) or
	    die "Cannot open $newpath for write: $!\n";
	binmode(PTWFILE);
	my $buffer;
	while (read($fh,$buffer,4096)) {
	    print PTWFILE $buffer;
	}
	close(PTWFILE)
	    or die "Error closing $newpath: $!\n";
	# Update metadata
	if ($mtime) {
	    utime $mtime,$mtime, $newpath;
	}
	if (defined $mode) {
	    chmod oct($mode) & 07777, $newpath;
	}
	# Move into final location
	rename($newpath,$dest)
	    or die "Failed to rename $newpath to $dest: $!\n";
    };
    if (my $exc = $@) {
	# Clean up temporary name
	unlink $newpath;
	die $exc;
    }
}

sub _writelink {
    my($self, $dir,$leaf,$link) = @_;
    my $dest = "$dir$leaf";
    my $newname = $self->tempname($dir);
    eval {
	symlink $link, $newname
	    or die "Cannot create symlink $newname: $!\n";
	rename $newname, "$dir/$leaf"
	    or die "Error renaming symlink as $dest: $!\n";
    };
    if (my $exc = $@) {
	# Clean up temporary name
	unlink $newname;
	die $exc;
    }
}

sub _mkdir {
    my($self, $dir,$leaf,$mode) = @_;
    my $newname = "$dir$leaf";
    mkdir($newname) or
	die "Failed to create directory: $!\n";
    if (defined $mode) {
	chmod oct($mode) & 07777, $newname;
    }
}

sub _remove {
    my($self, $dir,$leaf) = @_;
    my $dest = "$dir$leaf";
    File::Path::rmtree($dest);
}

#-----------------------------------------------------------------------
#	Check that user has correct permission.  If not, send an
#	appropriate response and return undef.
#-----------------------------------------------------------------------
sub permit {
    my($self, $query,$user,$action) = @_;
    my $permitted;
    my $list = $self->{"${action}Permission"};
    if (! defined($list)) {
	$permitted = undef;
    } elsif (ref($list) eq 'ARRAY') {
	$permitted = grep {$user eq $_} @$list;
    } elsif (ref($list) eq 'CODE') {
	$permitted = $list->($user);
    } elsif ($list eq '*') {
	$permitted = 1;
    } else {
	$permitted = undef;
    }
    if (! $permitted) {
	my $response = $query->response;
	my $html = "<title>Access denied</title>".
	    "You do not have $action access";
	if ($response->can('forbidden')) {
	    $response->forbidden($html);
	} elsif ($response->can('set_status')) {
	    $response->set_status(403); # Forbidden
	    $response->write($html);
	} else {
	    # Backward compatibility
	    my $q = $response->{'cgi'};	# FIXME naughty - need 403 method
	    print $q->header(-status=>403);
	    print $html;
	}
	return undef;
    } else {
	return 1;
    }
}

sub no_content {
    my($self, $response) = @_;
    if ($response->can('no_content')) {
	$response->no_content;
    } elsif ($response->can('set_status')) {
	$response->set_status('204 No Content');
	$response->write('');
    } else {
	# Backward compatibility
	my $q = $response->{'cgi'};	# FIXME naughty - need 404 method
	print $q->header(-status=>204);
    }
}

sub tempname {
    my($self, $dir) = @_;
    while (1) {
	my $name = sprintf('%s.webfiler.%08X',
			   $dir, time() ^ (rand(99)*19601120));
	return $name unless lstat $name;
	# Already exists - try another
    }
}

sub check_dir_and_type {
    my($dir,$leaf,@allowed) = @_;
    die "No such directory $dir\n"
	unless -d $dir;
    return &check_type("$dir$leaf", @allowed);
}

sub check_type {
    my($path,@allowed) = @_;
    my $type = &filetype($path);
    return $type if grep {$_ eq $type} @allowed;
    my $err = {
	'' => 'No such file or directory',
	d => 'Is a directory',
	l => 'Is a symlink',
	f => 'Is a file',
	'?' => 'Is a special file',
    }->{$type};
    die "$err: $path\n";
}

sub filetype {
    my($path) = @_;
    lstat($path) or return '';
    return 'l' if (-l _);
    return 'f' if (-f _);
    return 'd' if (-d _);
    return '?';
}

1;
